from sqlalchemy import *

from dataProviderBase import *
from T0Mon.utils import *
from T0Mon.analysis import *
from T0Mon.consts import *
from T0Mon.sqlManager import *

from T0Mon.globals import logger
import T0Mon.config

class RepackStatsDataProvider(StatsDataProvider):
    def __init__(self, runid, title = "Repack stats"):
        StatsDataProvider.__init__(self, title, TABLES.REPACK_JOBS)
        self.runid = runid
        StatsDataProvider.load(self)

    def _buildCols(self):
        self.cols = getStatsCols(self.tableName, self.runid)

    def _buildQuery(self, status):
        jobTable = Tbl(self.tableName)
        assocTable = Tbl(TABLES.JOB_DATASET_STREAMER_ASSOC)
        streamerTable = Tbl(TABLES.STREAMER)
        self.query = select(
            [func.count(distinct(jobTable.c.job_id))],
            and_(
                jobTable.c.job_status == status,
                streamerTable.c.run_id == self.runid
            ),
            from_obj = [
                jobTable.outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
                    outerjoin(streamerTable, assocTable.c.streamer_id == streamerTable.c.streamer_id)
            ]
        )

class MergeStatsDataProvider(StatsDataProvider):
    def __init__(self, runid, title = "Merge stats"):
        StatsDataProvider.__init__(self, title, TABLES.MERGE_JOBS)
        self.runid = runid
        StatsDataProvider.load(self)

    def _buildCols(self):
        self.cols = getStatsCols(self.tableName, self.runid)

    def _buildQuery(self, status):
        jobTable = Tbl(TABLES.MERGE_JOBS)
        assocTable = Tbl(TABLES.MERGE_JOB_REPACK_ASSOC)
        repackedTable = Tbl(TABLES.REPACKED)
        self.query = select(
            [func.count(distinct(jobTable.c.job_id))],
            and_(
                jobTable.c.job_status == status,
                repackedTable.c.run_id == self.runid
            ),
            from_obj = [
                jobTable.outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
                    outerjoin(repackedTable, assocTable.c.repacked_id == repackedTable.c.repacked_id)
            ]
        )
        # globals.logger.log(self.query)

class RecoStatsDataProvider(StatsDataProvider):
    def __init__(self, runid, title = "Reco stats"):
        StatsDataProvider.__init__(self, title, TABLES.RECO_JOBS)
        self.runid = runid
        StatsDataProvider.load(self)

    def _buildCols(self):
        self.cols = getStatsCols(self.tableName, self.runid)

    def _buildQuery(self, status):
        jobTable = Tbl(TABLES.RECO_JOBS)
        assocTable = Tbl(TABLES.RECO_JOB_REPACK_ASSOC)
        repackedTable = Tbl(TABLES.REPACKED)
        self.query = select(
            [func.count(distinct(jobTable.c.job_id))],
            and_(
                jobTable.c.job_status == status,
                repackedTable.c.run_id == self.runid
            ),
            from_obj = [
                jobTable.outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
                    outerjoin(repackedTable, assocTable.c.repacked_id == repackedTable.c.repacked_id)
            ]
        )
        # globals.logger.log(self.query)

class RepackMergeJobDP(PagedTableDataProvider):

    cols = Cols([
        Col("Job ID"),
        Col("Run ID", decorator=formRunLink),
        Col("Definition time", decorator=formatTime),
        Col("Repacked"),
        Col("Dataset")
    ])

    def __init__(self, orderby = 0, asc = True, page = 0, pageSize = config.PAGE_SIZE):
        PagedTableDataProvider.__init__(self, orderby, asc, page, pageSize = pageSize)
        self.title = "Merge jobs"

    def _buildCountQuery(self):
        jobTable = Tbl(TABLES.MERGE_JOBS)
        assocTable = Tbl(TABLES.MERGE_JOB_REPACK_ASSOC)
        repackTable = Tbl(TABLES.REPACKED)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)
        self.countQuery = select(
            [func.count(func.distinct(jobTable.c.job_id))],
            self.whereCond,
            from_obj = [
                jobTable.
                outerjoin(assocTable, jobTable.c.job_id == assocTable.c.job_id).
                outerjoin(repackTable, repackTable.c.repacked_id == assocTable.c.repacked_id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == repackTable.c.dataset_id)
            ]
        )

    def _buildMainQuery(self):
        jobTable = Tbl(TABLES.MERGE_JOBS)
        statusTable = Tbl(TABLES.JOB_STATUS)
        assocTable = Tbl(TABLES.MERGE_JOB_REPACK_ASSOC)
        repackTable = Tbl(TABLES.REPACKED)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)

        c = (
            jobTable.c.job_id,
            repackTable.c.run_id,
            jobTable.c.definition_time,
            func.count(assocTable.c.repacked_id).label("num_repacked"),
            datasetTable.c.name,
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            self.whereCond,
            from_obj = [
                jobTable.
                #outerjoin(statusTable).
                outerjoin(assocTable, jobTable.c.job_id == assocTable.c.job_id).
                outerjoin(repackTable, repackTable.c.repacked_id == assocTable.c.repacked_id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == repackTable.c.dataset_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                datasetTable.c.name,
                repackTable.c.run_id,
            ).order_by(o)


class RecoJobDP(PagedTableDataProvider):

    cols = Cols([
        Col("Job ID"),
        Col("Run ID", decorator=formRunLink),
        Col("Definition time", decorator=formatTime),
        Col("Merged"),
        Col("Dataset"),
    ])

    def __init__(self, orderby = 0, asc = True, page = 0, pageSize = config.PAGE_SIZE):
        PagedTableDataProvider.__init__(self, orderby, asc, page, pageSize = pageSize)

    def _buildCountQuery(self):
        jobTable = Tbl(TABLES.RECO_JOBS)
        assocTable = Tbl(TABLES.RECO_JOB_REPACK_ASSOC)
        repackTable = Tbl(TABLES.REPACKED)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)
        self.countQuery = select(
            [func.count(jobTable.c.job_id)],
            self.whereCond,
            from_obj = [
                jobTable.
                outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
                outerjoin(repackTable, repackTable.c.repacked_id == assocTable.c.repacked_id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == repackTable.c.dataset_id)
            ]
        )

    def _buildMainQuery(self):
        jobTable = Tbl(TABLES.RECO_JOBS)
        statusTable = Tbl(TABLES.JOB_STATUS)
        assocTable = Tbl(TABLES.RECO_JOB_REPACK_ASSOC)
        repackTable = Tbl(TABLES.REPACKED)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)

        c = (
             jobTable.c.job_id,
             repackTable.c.run_id,
             jobTable.c.definition_time,
             func.count(assocTable.c.repacked_id).label("num_merged"),
             datasetTable.c.name,
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            self.whereCond,
            from_obj = [
                jobTable.
                #outerjoin(statusTable).
                outerjoin(assocTable, assocTable.c.job_id == jobTable.c.job_id).
                outerjoin(repackTable, repackTable.c.repacked_id == assocTable.c.repacked_id).
                outerjoin(datasetTable, datasetTable.c.dataset_id == repackTable.c.dataset_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                datasetTable.c.name,
                repackTable.c.run_id,
            ).order_by(o)

class RepackJobDP(PagedTableDataProvider):

    cols = Cols([
        Col("Job ID"),
        Col("Run ID", decorator=formRunLink),
        Col("Definition time", decorator=formatTime),
        Col("Streamers"),
        Col("Stream"),
#        Col("Dataset")
    ])

    def __init__(self, orderby = 0, asc = True, page = 0, pageSize = config.PAGE_SIZE):
        PagedTableDataProvider.__init__(self, orderby, asc, page, pageSize = pageSize)
        self.title = "Repacker jobs"

    def _buildCountQuery(self):
        jobTable = Tbl(TABLES.REPACK_JOBS)
        assocTable = Tbl(TABLES.JOB_DATASET_STREAMER_ASSOC)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)
        streamerTable = Tbl(TABLES.STREAMER)
        self.countQuery = select(
            [func.count(func.distinct(jobTable.c.job_id))],
            self.whereCond,
            from_obj = [
                jobTable.
                outerjoin(assocTable, jobTable.c.job_id == assocTable.c.job_id).
                outerjoin(streamerTable, streamerTable.c.streamer_id == assocTable.c.streamer_id)
#                outerjoin(datasetTable, datasetTable.c.dataset_id == assocTable.c.dataset_id)
            ]
        )

    def _buildMainQuery(self):
        jobTable = Tbl(TABLES.REPACK_JOBS)
        assocTable = Tbl(TABLES.JOB_DATASET_STREAMER_ASSOC)
        statusTable = Tbl(TABLES.JOB_STATUS)
        datasetTable = Tbl(TABLES.PRIMARY_DATASET)
        streamerTable = Tbl(TABLES.STREAMER)

        c = (
             jobTable.c.job_id,
             streamerTable.c.run_id,
             jobTable.c.definition_time,
             func.count(func.distinct(assocTable.c.streamer_id)).label("num_streamers"),
             streamerTable.c.streamname,
#             datasetTable.c.name
        )

        if self.asc: o = c[self.orderby].asc()
        else: o = c[self.orderby].desc()

        self.mainQuery = select(
            c,
            self.whereCond,
            from_obj = [
                jobTable.
                outerjoin(assocTable, jobTable.c.job_id == assocTable.c.job_id).
                outerjoin(streamerTable, streamerTable.c.streamer_id == assocTable.c.streamer_id)
#                outerjoin(datasetTable, datasetTable.c.dataset_id == assocTable.c.dataset_id)
            ]
            ).group_by(
                jobTable.c.job_id,
                jobTable.c.definition_time,
                streamerTable.c.run_id,
                streamerTable.c.streamname,
#                datasetTable.c.name
            ).order_by(o)

class RepackJobByStreamerDP(RepackJobDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        RepackJobDP.__init__(self, orderby, asc, page)
        assocTable = Tbl(TABLES.JOB_DATASET_STREAMER_ASSOC)
        jobTable = Tbl(TABLES.REPACK_JOBS)
        self.whereCond = jobTable.c.job_id.in_(
            select([func.distinct(assocTable.c.job_id)], assocTable.c.streamer_id.in_(ids))
        )
        RepackJobDP._load(self)

class RepackJobByRepackIdsDP(RepackJobDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        RepackJobDP.__init__(self, orderby, asc, page)
        assocTable1 = Tbl(TABLES.JOB_DATASET_STREAMER_ASSOC)
        assocTable2 = Tbl(TABLES.REPACK_STREAMER_ASSOC)
        jobTable = Tbl(TABLES.REPACK_JOBS)
        self.whereCond = jobTable.c.job_id.in_(
            select([func.distinct(assocTable1.c.job_id)], assocTable1.c.streamer_id.in_(
                select([assocTable2.c.streamer_id], assocTable2.c.repacked_id.in_(ids))
            ))
        )
        RepackJobDP._load(self)


class RepackMergeJobByRepackIdsDP(RepackMergeJobDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        RepackMergeJobDP.__init__(self, orderby, asc, page)
        jobTable = Tbl(TABLES.MERGE_JOBS)
        assocTable = Tbl(TABLES.MERGE_JOB_REPACK_ASSOC)
        self.whereCond = jobTable.c.job_id.in_(
            select([func.distinct(assocTable.c.job_id)],
                assocTable.c.repacked_id.in_(ids))
        )
        RepackMergeJobDP._load(self)
        self.title = "Merger jobs"

class RecoJobByMergeIdsDP(RecoJobDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        RecoJobDP.__init__(self, orderby, asc, page)
        jobTable = Tbl(TABLES.RECO_JOBS)
        assocTable = Tbl(TABLES.RECO_JOB_REPACK_ASSOC)
        self.whereCond = jobTable.c.job_id.in_(
            select([assocTable.c.job_id], assocTable.c.repacked_id.in_(ids))
        )
        RecoJobDP._load(self)
        self.title = "Reco jobs"

class RecoJobByRecoIdsDP(RecoJobDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        RecoJobDP.__init__(self, orderby, asc, page)
        jobTable = Tbl(TABLES.RECO_JOBS)
        assocTable = Tbl(TABLES.RECO_JOB_REPACK_ASSOC)
        parentageTable = Tbl(TABLES.REPACKED_RECO_PARENTAGE)
        self.whereCond = jobTable.c.job_id.in_(
            select([assocTable.c.job_id], assocTable.c.repacked_id.in_(
                select([parentageTable.c.input_id], parentageTable.c.output_id.in_(ids))
            ))
        )
        RecoJobDP._load(self)
        self.title = "Reco jobs"

class RepackRunningJobDP(RepackJobDP):

    cols = Cols([
        Col("Job ID"),
        Col("Run ID", decorator=formRunLink),
        Col("Running time", decorator=deltaTime),
        Col("Streamers"),
        Col("Stream"),
        # ["Dataset", None]
    ])

    def __init__(self, orderby = 0, asc = True, page = 0):
        RepackJobDP.__init__(self, orderby, asc, page, pageSize = 25)
        jobTable = Tbl(TABLES.REPACK_JOBS)
        self.whereCond = or_(
            jobTable.c.job_status == JOB_USED,
            jobTable.c.job_status == JOB_NEW,
        )
        RepackJobDP._load(self)
        self.title = "Repack running jobs"

class RepackMergeRunningJobDP(RepackMergeJobDP):

    cols = Cols([
        Col("Job ID"),
        Col("Run ID", decorator=formRunLink),
        Col("Running time", decorator=deltaTime),
        Col("Repacked"),
        Col("Dataset")
    ])

    def __init__(self, orderby = 0, asc = True, page = 0):
        RepackMergeJobDP.__init__(self, orderby, asc, page, pageSize = 25)
        jobTable = Tbl(TABLES.MERGE_JOBS)
        self.whereCond = or_(
            jobTable.c.job_status == JOB_USED,
            jobTable.c.job_status == JOB_NEW,
        )
        RepackMergeJobDP._load(self)
        self.title = "Merge running jobs"

class RecoRunningJobDP(RecoJobDP):

    cols = Cols([
        Col("Job ID"),
        Col("Run ID", decorator=formRunLink),
        Col("Reco time", decorator=deltaTime),
        Col("Merged"),
        Col("Dataset"),
    ])

    def __init__(self, orderby = 0, asc = True, page = 0):
        RecoJobDP.__init__(self, orderby, asc, page, pageSize = 25)
        jobTable = Tbl(TABLES.RECO_JOBS)
        self.whereCond = or_(
            jobTable.c.job_status == JOB_USED,
            jobTable.c.job_status == JOB_NEW,
        )
        RecoJobDP._load(self)
        self.title = "Reco running jobs"

class RepackJobByStatusDP(RepackJobDP):

    def __init__(self, status, orderby = 0, asc = True, page = 0):
        RepackJobDP.__init__(self, orderby, asc, page)
        jobTable = Tbl(TABLES.REPACK_JOBS)
        self.whereCond = jobTable.c.job_status == status
        RepackJobDP._load(self)
        s = JOB_STATUS[status]
        self.title = "Repacker " + s + " jobs"

class RepackMergeJobByStatusDP(RepackMergeJobDP):

    def __init__(self, status, orderby = 0, asc = True, page = 0):
        RepackMergeJobDP.__init__(self, orderby, asc, page)
        jobTable = Tbl(TABLES.MERGE_JOBS)
        self.whereCond = jobTable.c.job_status == status
        RepackMergeJobDP._load(self)
        s = JOB_STATUS[status]
        self.title = "Merge " + s + " jobs"

class RecoJobByStatusDP(RecoJobDP):

    def __init__(self, status, orderby = 0, asc = True, page = 0):
        RecoJobDP.__init__(self, orderby, asc, page)
        jobTable = Tbl(TABLES.RECO_JOBS)
        self.whereCond = jobTable.c.job_status == status
        RecoJobDP._load(self)
        s = JOB_STATUS[status]
        self.title = "Reco " + s + " jobs"

class RepackJobByStatusAndRunIdDP(RepackJobDP):

    def __init__(self, status, runid, orderby = 0, asc = True, page = 0):
        RepackJobDP.__init__(self, orderby, asc, page)
        streamerTable = Tbl(TABLES.STREAMER)
        jobTable = Tbl(TABLES.REPACK_JOBS)
        self.whereCond = and_(
            jobTable.c.job_status == status,
            streamerTable.c.run_id == runid,
        )
        RepackJobDP._load(self)
        s = JOB_STATUS[status]
        self.title = "Repacker " + s + " jobs for run " + str(runid)

class RepackMergeJobByRepackMergeIdsDP(RepackMergeJobDP):
    def __init__(self, ids, orderby = 0, asc = True, page = 0):
        RepackMergeJobDP.__init__(self, orderby, asc, page)
        assocTable = Tbl(TABLES.MERGE_JOB_REPACK_ASSOC)
        jobTable = Tbl(TABLES.MERGE_JOBS)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE)
        self.whereCond = jobTable.c.job_id.in_(
            select([assocTable.c.job_id], assocTable.c.repacked_id.in_(
                select([parentageTable.c.input_id], parentageTable.c.output_id.in_(ids))
            )
        ))
        RepackMergeJobDP._load(self)

class RepackMergeJobByStatusAndRunIdDP(RepackMergeJobDP):

    def __init__(self, status, runid, orderby = 0, asc = True, page = 0):
        RepackMergeJobDP.__init__(self, orderby, asc, page)
        repackedTable = Tbl(TABLES.REPACKED)
        jobTable = Tbl(TABLES.MERGE_JOBS)
        self.whereCond = and_(
            jobTable.c.job_status == status,
            repackedTable.c.run_id == runid,
        )
        RepackMergeJobDP._load(self)
        s = JOB_STATUS[status]
        self.title = "Merger " + s + " jobs for run " + str(runid)

class RecoJobByStatusAndRunIdDP(RecoJobDP):

    def __init__(self, status, runid, orderby = 0, asc = True, page = 0):
        RecoJobDP.__init__(self, orderby, asc, page)
        repackedTable = Tbl(TABLES.REPACKED)
        jobTable = Tbl(TABLES.RECO_JOBS)
        self.whereCond = and_(
            jobTable.c.job_status == status,
            repackedTable.c.run_id == runid,
        )
        RecoJobDP._load(self)
        s = JOB_STATUS[status]
        self.title = "Reco " + s + " jobs for run " + str(runid)
